/*
 * NAME:
 *	HC - Hanzi Converter Version 1.2u
 *	Copyright @ 1988,1989,1990 by Fung F. Lee & Ricky Yeung
 *
 * SYNOPSIS:
 *	HC -m mode [-tabs t1 t2 ...]
 *		mode: can be BIGtoGB or GBtoBIG
 *		ti: conversion table filenames
 *
 * DESCRIPTION:
 *	`HC' converts BIG-5 codes into GB codes (with high bit set),
 *	and vice versa. BIG-5 is a popular hanzi code used in Taiwan,
 *	whereas GB (GuoBiao) is Mainland China's national standard.
 *	GB code (with high bit set) is adopted widely in many systems,
 *	such as CCDOS and Macintosh Chinese Operating System 6.0.3.
 *	
 * FILES:
 *	GB0toBIG, GB1toBIG, GB2toBIG, BIG0toGB, BIG1toGB, YiTiZi
 *	are HC mapping table files. They are either GB or BIG-5 files.
 *	The format is simple and allows easy modification with a Chinese
 *	editor. 
 *
 * AUTHORS:
 *	Ricky Yeung (yeung@june.cs.washington.edu)
 *	Fung F. Lee (lee@milo.stanford.edu)
 *
 * ACKNOWLEDGEMENT:
 *	Thanks to Edmund Lai for providing most of the data in `GB2toBIG'.
 *
 * DISTRIBUTION:
 *	This program and the data files are NOT in public domain.  
 *	The source file and the related table files as listed
 *	in FILES can be freely distributed for non-commercial purposes only.
 *
 *	Contact the authors before making any modification to the software
 *	and/or the table files.
 *
 *	This software is provided "as is" and the authors disclaim all
 *	warranties with regard to this software and the correctness of the
 *	mapping table files.  
 *
 */

#include <stdio.h>
#include <ctype.h>

#define	GBtoBIG	1
#define BIGtoGB	2
#define YITIZI	3

#define GBstart 0xA1A1	/* first address of GtoB table */
#define GBend	0xFEFE	/* last address of GtoB table */
#define GBsize	0x5E5E	/* size of GtoB table */
#define BIGstart 0xA140	/* first address of BtoG table */
#define BIGend	0xF9FE	/* last address of BtoG table */
#define	BIGsize	0x58BF	/* size of BtoG table */

#define GBbox	0xA1F5	/* GB code of blank box, default value of BtoG table */
#define BIGbox	0xA1BC	/* BIG code of blank box,default value of GtoB table */

unsigned int *BtoG, *GtoB;	/* code mapping tables */

#define HB(x)		((x>>8) & 0xFF)
#define LB(x)		(x & 0xFF)
#define DB(hi,lo)	(((hi)&0xFF) << 8 | (lo)&0xFF)
#define SGBtoGB(hi,lo)	(((hi)&0x7F) << 8 | (lo)&0x7F)
#define isFirstByte(c)	((c)>=0xA1)
#define inGBrange(x)	(((x)>=GBstart) && ((x)<=GBend))
#define inBIGrange(x)	(((x)>=BIGstart) && ((x)<=BIGend))

#define	SkipLine(fp)	while ((fgetc(fp)!='\n'));

int SkipSpaces(fp)
    FILE *fp;
{
    int c;
    
    for (;;)
    {
	c=fgetc(fp);
	if ((c&0x80) || !isspace(c)) break;
    }
    return(c);
}


fatal(s)
char *s;
{
    fprintf(stderr, "%s\n", s);
    exit(1);
}


FILE	*openfile(fn)
char *fn;
{
    FILE *fp = fopen(fn, "r");
    
    if (!fp)
    {
	fprintf(stderr, "Unable to open file: %s\n", fn);
	exit(1);
    }
    return(fp);
}


SetUpTables()
{
    int	i;
    
    GtoB = (unsigned int *) malloc(sizeof(int) * GBsize);
    BtoG = (unsigned int *) malloc(sizeof(int) * BIGsize);
    for (i=0; i<GBsize; i++)
	GtoB[i] = BIGbox;
    for (i=0; i<BIGsize; i++)
	BtoG[i] = GBbox;
}

#define	BUFSIZE	255

long  makeTables(fn)
char *fn;
{
    unsigned int codeS, codeT, cfirst, clast;
    int c1, c2, mode = GBtoBIG;
    long total = 0, makeYiTi();
    FILE *fp;
    char buf[BUFSIZE];
    
    fp = openfile(fn);
    
    for (;;)
    {
	fgets(buf, BUFSIZE, fp); /* no check for EOF */
	if (buf[0]!='\n' && (buf[0]!='#') && (buf[0]!=' ')) break;
    }

    /* header in new HCF format: e.g. '%GB->BIG', '%BIG->GB', '%YITIZI' */
    if (buf[0] == '%')
    {
	switch(buf[1])
	{
	  case 'G': mode = GBtoBIG; break;
	  case 'B': mode = BIGtoGB; break;
	  case 'Y': mode = YITIZI; break;
	  default: fatal("unknown HCF format\n"); break;
	}
    }
    
    if (mode == YITIZI)
	return makeYiTi(fp);
    
    for (;;)
    {
	if (fscanf(fp, "%x %x", &cfirst, &clast) != 2)
	    break;
	for (codeS = cfirst; codeS <= clast; codeS++)
	{
	    if ((c1=SkipSpaces(fp))==EOF) break;
	    c2 = fgetc(fp);
	    codeT = DB(c1, c2);
	    if (mode == GBtoBIG)
	    {
		if (!inBIGrange(codeT))
		    fatal("error: invalid code mapping table entry\n");
		if (codeT != BIGbox)
		{
		    GtoB[codeS - GBstart] = codeT;
		    BtoG[codeT - BIGstart] = codeS;
		    total++;
		}
	    }
	    else if (mode == BIGtoGB)
	    {
		if (!inGBrange(codeT))
		    fatal("error: invalid code mapping table entry\n");
		if (codeT != GBbox)
		{
		    BtoG[codeS - BIGstart] = codeT;
		    GtoB[codeT - GBstart] = codeS;
		    total++;
		}
	    }
	}
    }
    return total;
}


long makeYiTi(fp)
FILE *fp;
{
    unsigned int code1, code2, value;
    int		c1, c2;
    long	total=0;
    
    for (;;)
    {
	if ((c1=SkipSpaces(fp))==EOF) break;
	c2 = fgetc(fp);
	code1 = DB(c1, c2);
	value = BtoG[code1 - BIGstart];
	for (;;)
	{
	    if ((c1=SkipSpaces(fp))==EOF) break;
	    c2 = fgetc(fp);
	    code2 = DB(c1, c2);
	    BtoG[code2 - BIGstart] = value;
	    total++;		
	    if ((c1=fgetc(fp))=='\n')
		break;
	    ungetc(c1, fp);
	}
    }
    return total;
}


long converter(mode)
int mode;
{
    int c1, c2;
    unsigned int codeS, codeT;
    long unconverted = 0L;
    
    while ((c1=getchar())!=EOF)
    {
	if (!isFirstByte(c1))
	{
	    putchar(c1);
	}
	else
	{
	    c2 = getchar();
	    codeS = DB(c1, c2);
	    if (mode == GBtoBIG)
	    {
		if (inGBrange(codeS)) 
		    codeT = GtoB[codeS - GBstart];
		if ((codeT == BIGbox) && (codeS != GBbox))
		    unconverted++;
	    }
	    else if (mode == BIGtoGB)
	    {
		if (inBIGrange(codeS))
		    codeT = BtoG[codeS - BIGstart];
		if ((codeT == GBbox) && (codeS != BIGbox))
		    unconverted++;
	    }
	    putchar(HB(codeT));
	    putchar(LB(codeT));
	}
    }
    return unconverted;
}


usage(s)
char *s;
{
    fprintf(stderr, "usage: %s -m mode [-tabs t1 t2 ...]\n", s);
    fprintf(stderr, "\t\t\tmode can be BIGtoGB or GBtoBIG\n");
    fprintf(stderr, "\t\t\tti is conversion table filename\n");
    exit(1);
}


main(argc, argv)
int argc;
char *argv[];
{
    int	mode = 0, i;
    char *pname;
    long miss;
    
    pname = argv[0];
    SetUpTables();
    for( i=1; i<argc; i++)
    {
	if ((!strcmp(argv[i], "-mode")) || (!strcmp(argv[i], "-m")))
	{
	    i++;
	    if (!strcmp(argv[i], "BIGtoGB"))
		mode = BIGtoGB;
	    else if (!strcmp(argv[i], "GBtoBIG"))
		mode = GBtoBIG;
	    else usage(pname);
	}
	else if (!strcmp(argv[i], "-tabs"))
	    while (1)
	    {
		if (++i >= argc) break;
		if (argv[i][0]=='-') /* other option, put it back */
		{
		    i--;
		    break;
		}
		else makeTables(argv[i]);
	    }
	else usage(pname);
    }
    if ((mode!=BIGtoGB)&&(mode!=GBtoBIG)) usage(pname);
    miss = converter(mode); 
    if (miss > 0)
	fprintf(stderr, "number of hanzi not converted = %ld\n", miss);
}
